{ -------------------------------------------------------------------------------------}
{ A "Speedy TreeView" component for Delphi32.                                          }
{ Copyright 1997, Hkon Eines.  All Rights Reserved.                                   }
{ This component can be freely used and distributed in commercial and private          }
{ environments, provided this notice is not modified in any way.                       }
{ -------------------------------------------------------------------------------------}
{ Feel free to contact me if you have any questions, comments or suggestions at        }
{   haakon.eines@finale.no (Hkon Eines)                                               }
{ You can always find the latest version of this component at:                         }
{   http://www.worldnet.net/~cycocrew/delphi/                                          }
{ -------------------------------------------------------------------------------------}
{ Date last modified:  05/05/97                                                        }
{ -------------------------------------------------------------------------------------}

{ -------------------------------------------------------------------------------------}
{ THETreeView v1.02                                                                    }
{ -------------------------------------------------------------------------------------}
{ Description:                                                                         }
{   A component that enhances the TTreeView component                                  }
{ Properties:                                                                          }
{   property PrintOutTitle: string;                                                    }
{   property SortType: TSortType;                                                      }
{ Procedures and functions:                                                            }
{   procedure Print;                                                                   }
{ -------------------------------------------------------------------------------------}
{ Revision History:                                                                    }
{ 1.00:  + First release                                                               }
{ 1.01:  + Removed tree2vw from uses clause (not needed)                               }
{ 1.02:  + Procedure SetItemList now takes care of Objects                             }
{        + Added procedure Print (Output of treenodes to printer)                      }
{        + property PrintOutTitle (Title of document when printing)                    }
{        + Automatic Scrolling up/down when dragging an item                           }
{--------------------------------------------------------------------------------------}

unit HETreeView;
{$R-}

(*
  TTREEVIEW:
    128 sec. to load 1000 items (no sorting)*
    270 sec. to save 1000 items (4.5 minutes!!!)

  THETREEVIEW:
    1.5 sec. to load 1000 items - about 850% faster!!! (2.3 seconds with sorting = stText)* 
    0.7 sec. to save 1000 items - about 3850% faster!!!

  NOTES:
  - All timings performed on a slow 486SX 33 MhZ, 20 Mb RAM.

  - * If the treeview is empty, loading takes 1.5 seconds,
    else add 1.5 seconds to clear 1000 items (a total loading time of 3 seconds).
    This is also the case for the TTreeView component (a total of 129.5 seconds).
    The process of clearing the items, is a call to
    SendMessage(hwnd, TVM_DELETEITEM, 0, Longint(TVI_ROOT)).
*)

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics,
  Controls, Forms, Dialogs, ComCtrls, CommCtrl, Printers;

type
  THETreeView = class(TTreeView)
  private
    FSortType: TSortType;
    FPrintOutTitle: string;
    procedure SetSortType(Value: TSortType);
  protected
    function GetItemText(ANode: TTreeNode): string;
    procedure DragOver(Source: TObject; X, Y: Integer; State: TDragState;
      var Accept: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    function AlphaSort: Boolean;
    function CustomSort(SortProc: TTVCompare; Data: Longint): Boolean;
    procedure LoadFromFile(const AFileName: string);
    procedure SaveToFile(const AFileName: string);
    procedure GetItemList(AList: TStrings);
    procedure SetItemList(AList: TStrings);
    //'Bold' should have been a property of TTreeNode, but...
    function IsItemBold(ANode: TTreeNode): Boolean;
    procedure SetItemBold(ANode: TTreeNode; Value: Boolean);
    procedure Print;
  published
    property PrintOutTitle: string read FPrintOutTitle write FPrintOutTitle;
    property SortType: TSortType read FSortType write SetSortType default stNone;
  end;

  procedure Register;

implementation

function DefaultTreeViewSort(Node1, Node2: TTreeNode; lParam: Integer): Integer; stdcall;
begin
  {with Node1 do
    if Assigned(TreeView.OnCompare) then
      TreeView.OnCompare(Node1.TreeView, Node1, Node2, lParam, Result)
    else}
  Result := AnsiCompareText(Node1.Text, Node2.Text);
end;

constructor THETreeView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FSortType := stNone;
  FPrintOutTitle := '';
end;

procedure THETreeView.SetItemBold(ANode: TTreeNode; Value: Boolean);
var
  Item: TTVItem;
  Template: Integer;
begin
  if ANode = nil then Exit;

  if Value then Template := -1
  else Template := 0;
  with Item do
  begin
    mask := TVIF_STATE;
    hItem := ANode.ItemId;
    stateMask := TVIS_BOLD;
    state := stateMask and Template;
  end;
  TreeView_SetItem(Handle, Item);
end;

function THETreeView.IsItemBold(ANode: TTreeNode): Boolean;
var
  Item: TTVItem;
begin
  Result := False;
  if ANode = nil then Exit;

  with Item do
  begin
    mask := TVIF_STATE;
    hItem := ANode.ItemId;
    if TreeView_GetItem(Handle, Item) then
      Result := (state and TVIS_BOLD) <> 0;
  end;
end;

procedure THETreeView.SetSortType(Value: TSortType);
begin
  if SortType <> Value then
  begin
    FSortType := Value;
    if ((SortType in [stData, stBoth]) and Assigned(OnCompare)) or
      (SortType in [stText, stBoth]) then
      AlphaSort;
  end;
end;

procedure THETreeView.LoadFromFile(const AFileName: string);
var
  AList: TStringList;
begin
  AList := TStringList.Create;
  Items.BeginUpdate;
  try
    AList.LoadFromFile(AFileName);
    SetItemList(AList);
  finally
    Items.EndUpdate;
    AList.Free;
  end;
end;

procedure THETreeView.SaveToFile(const AFileName: string);
var
  AList: TStringList;
begin
  AList := TStringList.Create;
  try
    GetItemList(AList);
    AList.SaveToFile(AFileName);
  finally
    AList.Free;
  end;
end;

procedure THETreeView.SetItemList(AList: TStrings);
var
  ALevel, AOldLevel, i, Cnt: Integer;
  S: string;
  ANewStr: string;
  AParentNode: TTreeNode;
  TmpSort: TSortType;

  function GetBufStart(Buffer: PChar; var ALevel: Integer): PChar;
  begin
    ALevel := 0;
    while Buffer^ in [' ', #9] do
    begin
      Inc(Buffer);
      Inc(ALevel);
    end;
    Result := Buffer;
  end;

begin
  //Delete all items - could have used Items.Clear (almost as fast)
  SendMessage(handle, TVM_DELETEITEM, 0, Longint(TVI_ROOT));
  AOldLevel := 0;
  AParentNode := nil;

  //Switch sorting off
  TmpSort := SortType;
  SortType := stNone;
  try
    for Cnt := 0 to AList.Count-1 do
    begin
      S := AList[Cnt];
      if (Length(S) = 1) and (S[1] = Chr($1A)) then Break;

      ANewStr := GetBufStart(PChar(S), ALevel);
      if (ALevel > AOldLevel) or (AParentNode = nil) then
      begin
        if ALevel - AOldLevel > 1 then raise Exception.Create('Invalid TreeNode Level');
      end
      else begin
        for i := AOldLevel downto ALevel do
        begin
          AParentNode := AParentNode.Parent;
          if (AParentNode = nil) and (i - ALevel > 0) then
            raise Exception.Create('Invalid TreeNode Level');
        end;
      end;
      AParentNode := Items.AddChildObject(AParentNode, ANewStr, AList.Objects[Cnt]);
      AOldLevel := ALevel;
    end;
  finally
    //Switch sorting back to whatever it was...
    SortType := TmpSort;
  end;
end;

procedure THETreeView.GetItemList(AList: TStrings);
var
  i, Cnt: integer;
  ANode: TTreeNode;
begin
  AList.Clear;
  Cnt := Items.Count -1;
  ANode := Items.GetFirstNode;
  for i := 0 to Cnt do
  begin
    AList.Add(GetItemText(ANode));
    ANode := ANode.GetNext;
  end;
end;

function THETreeView.GetItemText(ANode: TTreeNode): string;
begin
  Result := StringOfChar(' ', ANode.Level) + ANode.Text;
end;

function THETreeView.AlphaSort: Boolean;
begin
  if HandleAllocated then
  begin
    Result := CustomSort(nil, 0);
  end
  else Result := False;
end;

function THETreeView.CustomSort(SortProc: TTVCompare; Data: Longint): Boolean;
var
  SortCB: TTVSortCB;
  Node: TTreeNode;
begin
  Result := False;
  if HandleAllocated then
  begin
    with SortCB do
    begin
      if not Assigned(SortProc) then lpfnCompare := @DefaultTreeViewSort
      else lpfnCompare := SortProc;
      hParent := TVI_ROOT;
      lParam := Data;
      Result := TreeView_SortChildrenCB(Handle, SortCB, 0);
    end;

    if Items.Count > 0 then
    begin
      Node := Items.GetFirstNode;
      while Node <> nil do
      begin
        if Node.HasChildren then Node.CustomSort(SortProc, Data);
        Node := Node.GetNext;
      end;
    end;
  end;
end;

procedure THETreeView.DragOver(Source: TObject; X, Y: Integer; State: TDragState;
                   var Accept: Boolean);
begin
  inherited DragOver(Source, X, Y, State, Accept);
  if Accept then
  begin
    //Scroll up/down when dragging an item
    if (Y > Height-20) and (Y < Height) then
      PostMessage(Handle, WM_VSCROLL, MakeLong(SB_LINEDOWN, 0), 0); //UP
    if (Y < 20) then
      PostMessage(Handle, WM_VSCROLL, MakeLong(SB_LINEUP, 0), 0);   //DOWN
  end;
end;

procedure THETreeView.Print;
var
  AFile: TextFile;
  i, Cnt: integer;
  ANodeStr: string;
  ANode: TTreeNode;
begin
  //No Default Printer?
  if Printer.Printers.Count <= 0 then
    raise Exception.Create('No default printer selected');

  if Items.Count <= 0 then
    raise Exception.Create('Nothing to print');

  AssignPrn(AFile);
  try
    Rewrite(AFile);
    if Trim(FPrintOutTitle) <> '' then
    begin
      with Printer.Canvas.Font do
      begin
        Name  := 'Arial';
        Style := [fsBold];
        Size  := 14;
      end;
      WriteLn(AFile, FPrintOutTitle);
      Writeln(AFile, '');
    end;

    with Printer.Canvas.Font do
    begin
      Name  := 'Courier New';
      Style := [];
      Size  := 10;
    end;

    Cnt := Items.Count -1;
    ANode := Items[0];
    for i := 0 to Cnt do
    begin
      ANodeStr := StringOfChar(' ', 2*ANode.Level) + ANode.Text;
      Writeln(AFile, ANodeStr);
      ANode := ANode.GetNext;
    end;
  finally
    System.CloseFile(AFile);
  end;
end;

//Component Registration
procedure Register;
begin
  RegisterComponents('Win32', [THETreeView]);
end;


end.
